#!/bin/bash

export SCRIPTDIR="$(builtin cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"

declare -Ax __cached

############## HELPER FUNCTIONS ##############

LEGACY_GIT="$(git --version | grep 'git version 1')"

function die {
	log_error "$1"; exit 1;
}

function branch_remote_url {
    if [ "$LEGACY_GIT" ]; then
        echo -n "$(git remote -v | grep '(fetch)' | sed -E -e 's/(.+)\s+(.+)\s+(.+)/\2/')"
    else
        git remote get-url "$(git config branch."$(git symbolic-ref --short HEAD)".remote)" 2> /dev/null
    fi
}

function latest_merged_tag {
    if [ "$LEGACY_GIT" ]; then
        git for-each-ref --sort=-taggerdate refs/tags | cut -f2 |
            while read tag; do
                if [ -z "$(basename "$tag" | filter_tags "$1")" ]; then
                    continue
                fi
                if [ "$(git merge-base "$tag" HEAD)"="$(git rev-parse "$tag"^{commit})" ]; then
                    echo "$(basename "$tag")"
                    return
                fi
            done
    else
        git tag --list --sort=-taggerdate "$1*" --merged 2> /dev/null | filter_tags "$1" | head -n 1
    fi
}

function merged_tags {
    if [ "$LEGACY_GIT" ]; then
        git for-each-ref --sort=taggerdate refs/tags | cut -f2 |
            while read tag; do
                if [ -z "$(basename "$tag" | filter_tags "$1")" ]; then
                    continue
                fi
                if [ "$(git merge-base "$tag" HEAD)"="$(git rev-parse "$tag"^{commit})" ]; then
                    echo "$(basename "$tag")"
                fi
            done
    else
        git tag --list --sort=taggerdate "$1*" --merged 2> /dev/null | filter_tags "$1"
    fi
}

function formatted_changelog_record {
    declare data= name= tag= header_locale= header_date_format= body_wrap= "$@"

    verrel="$(echo "$tag" | sed -E -n "s/^$name-([^-]+)(-[^-]+)$/\1\2/p")"

    if [ "$LEGACY_GIT" ]; then
        if [ -n "$header_date_format" ]; then
             realdate="$(git for-each-ref "refs/tags/$tag" --format '%(taggerdate:iso8601)')"
             header_date="$(date --date "$realdate" +"$header_date_format")"
        fi
        header="$(LC_ALL="$header_locale" git for-each-ref "refs/tags/$tag" --format="* $header_date %(taggername) %(taggeremail)") $verrel"
    else
        if [ -n "$header_date_format" ]; then
            header_date_format_suffix=":format:$header_date_format"
        else
            header_date_format_suffix=
        fi
        header="$(LC_ALL="$header_locale" git tag -l "$tag" --format="* %(taggerdate$header_date_format_suffix) %(taggername) %(taggeremail)") $verrel"
    fi

    formatted_data="$(echo "$data" | fold -cs --width="$body_wrap" | sed -E -e 's/(\s+)%([^%])/\1%%\2/g' -e 's/\s*$//')"
    echo -n "${header}"$'\n'"${formatted_data}"
}

function encode_decimal {
    radixdigits=({a..z} {0..9})
    x="$1"
    result=
    while [ "$x" -gt 0 ]; do
        div=$(($x/${#radixdigits[@]}))
        mod=$(($x-$div*${#radixdigits[@]}))
        result="${radixdigits[$mod]}${result}"
        x="$div"
    done
    echo -n "$result"
}

function git_latest_ctime {
    mapfile -t files < <(git status --porcelain | cut -c 4- | sed -E -e 's|(.*) -> (.*)|\1\n\2|')
    latest_ctime=0
    for f in "${files[@]}"; do
        if [ -e "$(git rev-parse --show-toplevel)/$f" ]; then
            ctime=$(stat --format='%Z' "$(git rev-parse --show-toplevel)/$f")
            if [ "$ctime" -gt "$latest_ctime" ]; then
                latest_ctime="$ctime"
            fi
        fi
    done
    echo -n "$latest_ctime"
}

function path_to_name_suffix {
    sed -e 's|//*|-|g' -e 's|^\([^-]\)|-\1|' -e 's|-$||'
}

function filter_tags {
    grep -E "^$1-[^-]+-[^-]+$"
}

function output {
    __cached["${FUNCNAME[1]}"]="$*"
    echo -n "${__cached[${FUNCNAME[1]}]}"
}

function cached {
    if [ -n "${__cached[$1]}" ]; then
        echo -n "${__cached[$1]}"
    fi
}

function log_debug {
    log "DEBUG" "$@"
}

function log_info {
    log "INFO" "$@"
}

function log_error {
    log "ERROR" "$@"
}

function log {
    PREFIX=$1
    shift

    regexp="^\s*$"
    if [[ "$*" =~ $regexp ]]; then
        return
    fi

    if [ "${FUNCNAME[@]: -1}" != "main" ]; then
        echo "$PREFIX: ${FUNCNAME[@]: -1}: $*"
    else
        echo "$PREFIX: ${FUNCNAME[@]: -2}: $*"
    fi > /dev/stderr
}

function cached_git_name_version {
    echo -n "$(cached git_name)-$(cached git_version)"
}

############## REPO FUNCTIONS ##############

function git_name {
    declare name= prepend= append= "$@"

    if [ -n "$name" ]; then
        output "${prepend}${name}${append}"
        return
    fi

    if ! git rev-parse --show-toplevel &> /dev/null; then
        log_error "Not a Git repository."
        return 1
    fi

    url="$(branch_remote_url)"

    if [ -z "$url" ]; then
        url="$(git config --get remote.origin.url)"
    fi

    if [ -z "$url" ]; then
        log_error "Could not get remote URL."
        return 2
    fi

    url_last_part="$(basename "$url")"
    name="${url_last_part%.git}"

    output "${prepend}${name}${append}"
}

function git_bumped_version {
    declare lead=0 follow= latest_tag_version= "$@"

    if [ -n "$follow" ]; then
        echo -n "$lead.$follow"
        return
    fi

    follow="$(echo "$latest_tag_version" | sed -E -n "s/^$lead\.([^.]*)$/\1/p")"

    if [ -z "$follow" ] && [ "$lead" -ne 0 ]; then
        echo -n "$lead.0"
        return
    fi

    if echo "$follow" | grep -q '[^0-9]'; then
        log_error "$follow is not a number. Please, bump version manually."
        return 1
    fi

    bumped_follow=$(( $follow + 1 ))
    echo -n "$lead.$bumped_follow"
}

function git_version {
    declare name="$(cached git_name)" lead=0 follow= "$@"

    if [ -z "$name" ]; then
        log_error "name cannot be empty."
        return 1
    fi

    if echo "$lead.$follow" | grep -q '-'; then
        log_error "lead and follow cannot contain dashes."
        return 1
    fi

    if echo "$follow" | grep -q '\.'; then
        log_error "follow cannot contain dots."
        return 1
    fi

    latest_tag="$(latest_merged_tag "$name")"
    latest_tag_version="$(echo $latest_tag | sed -E -n "s/^$name-([^-]+)-[^-]+$/\1/p")"

    if [ -n "$VERSION_BUMP" ]; then
        version="$(git_bumped_version lead="$lead" follow="$follow" latest_tag_version="$latest_tag_version")"

        retcode=$?
        if [ $retcode -ne 0 ]; then
            return $retcode
        fi

        output "$version"
        return
    fi

    if [ -n "$latest_tag" ]; then
        commit_count="$(git rev-list "$latest_tag"..HEAD | wc -l)"
    else
        commit_count="$(git rev-list HEAD 2> /dev/null | wc -l  || printf 0)"
    fi

    if [ "$commit_count" -eq 0 ]; then
        commit_count_appendix=
    else
        commit_count_appendix=".git.$commit_count.$(git rev-parse --short HEAD)"
    fi

    latest_ctime="$(git_latest_ctime)"
    if [ "$latest_ctime" -eq 0 ]; then
        wtree_appendix=
    else
        wtree_appendix=".wtree.$(encode_decimal "$latest_ctime")"
    fi

    if [ -z "$follow" ]; then
        follow="$(echo "$latest_tag_version" | sed -E -n "s/^$lead\.([^.]*)$/\1/p")"
    fi

    output "${lead}.${follow:-0}${commit_count_appendix}${wtree_appendix}"
}

function git_vcs {
    declare subtree= "$@"

    url="$(git config --get remote.origin.url)"

    if [ -z "$url" ]; then
        log_error "Start by setting origin URL."
        return 1
    fi

    ref="$(git rev-parse HEAD 2> /dev/null)"

    if [ -n "$ref" ]; then
        treeish_suffix="#$ref:$subtree"
    else
        treeish_suffix=
    fi

    output "git+${url}${treeish_suffix}"
}

function git_pack {
    declare path="$(git rev-parse --show-toplevel)" dir_name="$(cached_git_name_version)" source_name="$(cached_git_name_version).tar.gz" "$@"

    if [ -z "$OUTDIR" ]; then
        log_debug "OUTDIR is not set. No action taken."
        output "$source_name"
        return
    fi

    EXCLUDE_TMPFILE="$(mktemp)"
    STDERR_TMPFILE="$(mktemp)"

    find "$path" | git check-ignore --stdin > "$EXCLUDE_TMPFILE" 2> "$STDERR_TMPFILE"
    [ -s "$STDERR_TMPFILE" ] && die "Your git version does not support check-ignore"

    log_info "packing path $path"

    cmd_stdout="$($SCRIPTDIR/scripts/pack_sources.sh --exclude-from "$EXCLUDE_TMPFILE" "$path" "$dir_name" "$OUTDIR/$source_name")"
    cmd_retcode=$?

    rm "$EXCLUDE_TMPFILE"
    rm "$STDERR_TMPFILE"

    log_debug "$cmd_stdout"

    if [ $cmd_retcode -eq 0 ]; then
        log_info "Wrote: $OUTDIR/$source_name"
    else
        log_error "Error during source creation."
        return $cmd_retcode
    fi

    output "$source_name"
}

function git_archive {
    declare path="$(git rev-parse --show-toplevel)" dir_name="$(cached_git_name_version)" source_name="$(cached_git_name_version).tar.gz" "$@"

    if [ -z "$OUTDIR" ]; then
        log_debug "OUTDIR is not set. No action taken."
        output "$source_name"
        return
    fi

    if ! git rev-parse HEAD &> /dev/null; then
        log_error "No commits yet. Cannot archive the content."
        return 1
    fi

    if [ "$(git status --porcelain)" ]; then
        log_error "Your working tree is dirty. Commit first."
        return 2
    fi

    log_info "archiving $path:"$'\n'"$(git log -1 --decorate)"

    cmd_stdout="$(git -C "$path" archive --prefix="$dir_name/" -o "$OUTDIR/$source_name" HEAD)"
    cmd_retcode=$?

    log_debug "$cmd_stdout"

    if [ $cmd_retcode -eq 0 ]; then
        log_info "Wrote: $OUTDIR/$source_name"
    else
        log_error "Error during source creation."
        return $cmd_retcode
    fi

    output "$source_name"
}

function git_setup_macro {
    declare dir_name="$(cached_git_name_version)" "$@"
    output "%setup -q -n $dir_name"
}

function git_changelog {
    declare name="$(cached git_name)" since_tag= until_tag= header_locale=POSIX header_date_format="%a %b %d %Y" body_wrap=80 "$@"

    mapfile -t tags < <(merged_tags "$name")

    changelog=
    until_tag_hit=
    for tag in "${tags[@]}"; do
        if [ -n "$since_tag" ] && [ -z "$since_tag_hit" ] && [ "$since_tag" != "$tag" ]; then
            continue
        fi
        since_tag_hit=1

        data="$(git for-each-ref "refs/tags/$tag" --format='%(body)')"
        if [ -z "$data" ]; then
            data="(none)"
        fi

        formatted_record="$(formatted_changelog_record \
            data="$data" name="$name" tag="$tag" header_locale="$header_locale" header_date_format="$header_date_format" body_wrap="$body_wrap")"

        if [ -n "$changelog" ]; then
            changelog=$'\n\n'"$changelog"
        fi

        changelog="${formatted_record}${changelog}"

        if [ "$until_tag" = "$tag" ]; then
            break
        fi
    done

    output "$changelog"
}

############## DIR FUNCTIONS ##############

function git_dir_name {
    git_name append="$(git -C "$(dirname "$INPUT_PATH")" rev-parse --show-prefix | path_to_name_suffix)" "$@"
}

function git_dir_version {
    git_version "$@"
}

function git_dir_vcs {
    git_vcs subtree="$(git -C "$(dirname "$INPUT_PATH")" rev-parse --show-prefix)" "$@"
}

function git_dir_pack {
    git_pack path="$(dirname "$INPUT_PATH")" "$@"
}

function git_dir_archive {
    git_archive path="$(dirname "$INPUT_PATH")" "$@"
}

function git_dir_setup_macro {
    git_setup_macro "$@"
}

function git_dir_changelog {
    git_changelog "$@"
}

############## CURRENT WORKING DIR FUNCTIONS ##############

function git_cwd_name {
    git_name append="$(git -C "$(pwd)" rev-parse --show-prefix | path_to_name_suffix)" "$@"
}

function git_cwd_version {
    git_version "$@"
}

function git_cwd_vcs {
    git_vcs subtree="$(git -C "$(pwd)" rev-parse --show-prefix)" "$@"
}

function git_cwd_pack {
    git_pack path="$(pwd)" "$@"
}

function git_cwd_archive {
    git_archive path="$(pwd)" "$@"
}

function git_cwd_setup_macro {
    git_setup_macro "$@"
}

function git_cwd_changelog {
    git_changelog "$@"
}
